import bpy
import math
import mathutils

def get_addon_prefs():
    addon_name = __package__
    prefs = None
    try:
        prefs = bpy.context.preferences.addons[addon_name].preferences
    except KeyError:
        prefs = None
    return prefs

class HDRI_Rotator(bpy.types.Operator):
    """Rotate HDRI or Light Direction with your mouse"""
    bl_idname = "view3d.hdri_rotator"
    bl_label = "HDRI Rotator"
    bl_description = "Rotate HDRI or Light Direction with your mouse. Hold {SHIFT} for precision."
    bl_options = {'REGISTER', 'UNDO', 'BLOCKING'}

    
    def normalize_rotation_rate(self, delta_x, mode):
        """
        Normalize rotation rate across different HDRI control methods
        """
        # Add safety checks
        if delta_x == 0:
            return 0.0
            
        # Base constants for normalization
        BASE_RANGE = 360.0  # Full rotation range in degrees
        MOUSE_PIXELS_FOR_FULL_ROTATION = 1000.0  # Pixels needed for full rotation
        
        # Get user sensitivity preference (0.1 to 2.0 range)
        addon_prefs = get_addon_prefs()
        user_sensitivity = addon_prefs.sensitivity if addon_prefs else 0.8
        
        # Clamp sensitivity to valid range
        user_sensitivity = max(0.1, min(2.0, user_sensitivity))
        
        # Apply precision mode if shift is held
        precision_multiplier = 0.1 if self.shift_held else 1.0
        
        # Calculate base rotation rate
        base_rotation = (delta_x / MOUSE_PIXELS_FOR_FULL_ROTATION) * BASE_RANGE
        
        # Apply mode-specific scaling
        if mode == 'LIGHT_DIRECTION':
            # Light direction needs finer control due to 3D vector manipulation
            scale_factor = 0.25
        elif mode == 'HDRI_ROTATION':
            # Built-in studio lights (-180 to +180 range)
            scale_factor = 1.0
        elif mode == 'CUSTOM_HDRI_ROTATION':
            # Custom HDRIs with direct rotation control
            scale_factor = 1.0
        else:  # MAPPING_NODE_ROTATION
            # Mapping nodes (potentially going through math nodes)
            scale_factor = 1.0
        
        # Calculate final rotation with all factors
        final_rotation = (
            base_rotation * 
            scale_factor * 
            user_sensitivity * 
            precision_multiplier
        )
        
        return final_rotation

    def modal(self, context, event):
        try:
            if event.type == 'MOUSEMOVE':
                # Update shift state
                if event.shift != self.shift_held:
                    self.shift_held = event.shift
                    self.start_x = event.mouse_x
                    self.start_y = event.mouse_y
                    if self.mode == 'LIGHT_DIRECTION':
                        self.start_light_direction = self.scene.display.light_direction.copy()
                    else:
                        self.start_rotation = self.get_current_rotation()

                # Calculate mouse movement
                delta_x = event.mouse_x - self.start_x
                delta_y = event.mouse_y - self.start_y

                # Get normalized rotation based on mode
                rotation_change = self.normalize_rotation_rate(delta_x, self.mode)

                if self.mode == 'LIGHT_DIRECTION':
                    # Convert rotation change to radians for vector rotation
                    rot_x = (delta_y / delta_x) * rotation_change * math.pi / 180.0 if delta_x != 0 else 0
                    rot_z = -rotation_change * math.pi / 180.0
                    
                    new_light_direction = self.calculate_new_direction(
                        self.start_light_direction,
                        rot_x,
                        rot_z
                    )
                    self.scene.display.light_direction = new_light_direction
                
                elif self.mode == 'CUSTOM_HDRI_ROTATION':
                    if isinstance(self.start_rotation, (mathutils.Euler, mathutils.Vector)):
                        new_rotation = self.start_rotation.copy()
                        new_rotation.z = self.start_rotation.z + math.radians(rotation_change)
                    elif isinstance(self.start_rotation, (float, int)):
                        new_rotation = self.start_rotation + rotation_change
                    self.set_custom_hdri_rotation(new_rotation)
                
                elif self.mode == 'MAPPING_NODE_ROTATION':
                    if isinstance(self.start_rotation, (mathutils.Euler, mathutils.Vector)):
                        new_rotation = self.start_rotation.copy()
                        new_rotation.z = self.start_rotation.z + math.radians(rotation_change)
                    self.set_mapping_node_rotation(new_rotation)
                
                elif self.mode == 'HDRI_ROTATION':
                    new_rotation = self.start_rotation + rotation_change
                    # Wrap rotation to -180 to 180 degrees
                    new_rotation = ((new_rotation + 180) % 360) - 180
                    self.set_builtin_hdr_rotation(new_rotation)
                
                self.safe_redraw()
                return {'RUNNING_MODAL'}

            # Rest of the modal method remains unchanged
            elif event.type == 'RIGHTMOUSE' and event.value == 'RELEASE':
                self.cleanup()
                return {'FINISHED'}
            elif event.type in {'ESC'}:
                if self.mode == 'LIGHT_DIRECTION':
                    self.scene.display.light_direction = self.start_light_direction
                elif self.mode == 'CUSTOM_HDRI_ROTATION':
                    self.set_custom_hdri_rotation(self.start_rotation)
                elif self.mode == 'MAPPING_NODE_ROTATION':
                    self.set_mapping_node_rotation(self.start_rotation)
                elif self.mode == 'HDRI_ROTATION':
                    self.set_builtin_hdr_rotation(self.start_rotation)
                self.cleanup()
                return {'CANCELLED'}
            return {'RUNNING_MODAL'}

        except Exception as e:
            self.report({'ERROR'}, f"An error occurred: {e}")
            self.cleanup()
            return {'CANCELLED'}


    def invoke(self, context, event):
        try:
            # Store initial mouse position
            self.start_x = event.mouse_x
            self.start_y = event.mouse_y
            self.shift_held = event.shift

            # Store references to frequently accessed data
            self.window = context.window
            self.space = context.space_data
            self.scene = context.scene

            # Get a safe reference to the area
            self.area = self.get_area_safe()
            if not self.area:
                self.report({'WARNING'}, "Operator must be run in a 3D Viewport")
                return {'CANCELLED'}

            # Get shading information
            self.shading = self.space.shading
            self.shading_type = self.shading.type

            # Check if we're in Workbench with shadows
            is_workbench_shadows = (
                self.shading.show_shadows and  # Shadows are enabled
                ((self.shading_type == 'SOLID') or  # Either in Solid mode
                 (self.shading_type == 'RENDERED' and context.scene.render.engine == 'BLENDER_WORKBENCH'))  # Or Workbench Rendered
            )

            if is_workbench_shadows:
                # Workbench (Solid or Rendered) with shadows enabled; rotate light direction
                self.mode = 'LIGHT_DIRECTION'
                self.start_light_direction = self.scene.display.light_direction.copy()
                self.window.cursor_modal_set('SCROLL_XY')
            
            elif self.shading_type in {'MATERIAL', 'RENDERED'}:
                # Material Preview or Rendered modes (non-Workbench)
                if self.shading_type == 'MATERIAL':
                    self.use_scene_world = self.shading.use_scene_world
                else:  # RENDERED
                    self.use_scene_world = self.shading.use_scene_world_render

                if self.use_scene_world:
                    # Check for custom HDRI setup
                    self.rotation_input = self.find_rotation_input()
                    if self.rotation_input is not None:
                        # Custom HDRI with rotation input
                        default_value = self.rotation_input.default_value
                        if isinstance(default_value, (mathutils.Euler, mathutils.Vector)):
                            self.start_rotation = default_value.copy()
                        elif isinstance(default_value, (float, int)):
                            self.start_rotation = default_value
                        else:
                            self.report({'ERROR'}, "Unsupported rotation input type.")
                            return {'CANCELLED'}
                        self.mode = 'CUSTOM_HDRI_ROTATION'
                    else:
                        # Check for mapping node
                        self.mapping_node = self.find_mapping_node()
                        if self.mapping_node:
                            default_value = self.mapping_node.inputs['Rotation'].default_value
                            if isinstance(default_value, (mathutils.Vector, mathutils.Euler)):
                                self.start_rotation = default_value.copy()
                            else:
                                self.report({'ERROR'}, "Unsupported rotation input type in mapping node.")
                                return {'CANCELLED'}
                            self.mode = 'MAPPING_NODE_ROTATION'
                        else:
                            self.report({'WARNING'}, "No 'Rotation' input or Mapping node found in World nodes")
                            return {'CANCELLED'}
                else:
                    # Built-in HDRI
                    self.start_rotation = self.get_builtin_hdr_rotation()
                    self.mode = 'HDRI_ROTATION'

                # Set appropriate cursor for HDRI rotation
                self.window.cursor_modal_set('SCROLL_X')
            else:
                # Unsupported shading mode
                self.report({'WARNING'}, "Shading type not supported")
                return {'CANCELLED'}

            context.window_manager.modal_handler_add(self)
            return {'RUNNING_MODAL'}

        except Exception as e:
            self.report({'ERROR'}, f"Initialization failed: {str(e)}")
            self.cleanup()
            return {'CANCELLED'}


        
    def cleanup(self):
        # Reset cursor
        if hasattr(self, 'window') and self.window:
            self.window.cursor_modal_restore()
        
        # Clear all stored states
        for attr in ['shift_held', 'start_rotation', 'start_light_direction', 
                    'rotation_input', 'mapping_node', 'mode', 'start_x', 'start_y']:
            if hasattr(self, attr):
                delattr(self, attr)

    def safe_redraw(self):
        area = self.get_area_safe()
        if area:
            area.tag_redraw()

    def get_area_safe(self):
        # Safely get area reference
        for area in bpy.context.screen.areas:
            if area.type == 'VIEW_3D':
                return area
        return None

    def get_current_rotation(self):
        """Get current rotation value based on mode"""
        try:
            if self.mode == 'LIGHT_DIRECTION':
                return self.scene.display.light_direction.copy()
            elif self.mode == 'CUSTOM_HDRI_ROTATION':
                if isinstance(self.start_rotation, (mathutils.Euler, mathutils.Vector)):
                    return self.rotation_input.default_value.copy()
                elif isinstance(self.start_rotation, (float, int)):
                    return self.rotation_input.default_value
                return None  # Handle unexpected types
            elif self.mode == 'MAPPING_NODE_ROTATION':
                return self.mapping_node.inputs['Rotation'].default_value.copy()
            elif self.mode == 'HDRI_ROTATION':
                return self.get_builtin_hdr_rotation()
            return None  # Handle unexpected modes
        except Exception as e:
            self.report({'WARNING'}, f"Error getting rotation: {str(e)}")
            return None

    def find_rotation_input(self, node_tree=None, visited=None, depth=0):
        if visited is None:
            visited = set()
        if node_tree is None:
            world = self.scene.world
            if world is None or not world.use_nodes:
                return None
            node_tree = world.node_tree

        if node_tree in visited:
            return None
        visited.add(node_tree)

        for node in node_tree.nodes:
            if node.type == 'GROUP':
                if 'Rotation' in node.inputs:
                    rotation_input = node.inputs['Rotation']
                    if rotation_input.is_linked or rotation_input.default_value is not None:
                        return rotation_input
                # Recursively check inside the node group's node tree
                rotation_input = self.find_rotation_input(node.node_tree, visited, depth + 1)
                if rotation_input:
                    return rotation_input
            elif 'Rotation' in node.inputs:
                rotation_input = node.inputs['Rotation']
                if rotation_input.is_linked or rotation_input.default_value is not None:
                    return rotation_input
        return None

    def find_mapping_node(self):
        world = self.scene.world
        if world is None or not world.use_nodes:
            return None
        node_tree = world.node_tree

        visited_groups = set()
        nodes_to_check = list(node_tree.nodes)

        while nodes_to_check:
            node = nodes_to_check.pop()
            if node.type == 'GROUP':
                if node.node_tree and node.node_tree not in visited_groups:
                    visited_groups.add(node.node_tree)
                    nodes_to_check.extend(node.node_tree.nodes)
            elif node.type == 'MAPPING':
                # Check if this Mapping node is connected to an Environment Texture node
                if self.is_mapping_connected_to_env_texture(node):
                    return node
        return None

    def is_mapping_connected_to_env_texture(self, mapping_node):
        # Check downstream nodes to see if the mapping node ultimately connects to an Environment Texture
        nodes_to_check = [mapping_node]
        visited_nodes = set()

        while nodes_to_check:
            node = nodes_to_check.pop()
            visited_nodes.add(node)

            if node.type == 'TEX_ENVIRONMENT':
                return True

            for output in node.outputs:
                for link in output.links:
                    next_node = link.to_node
                    if next_node not in visited_nodes:
                        nodes_to_check.append(next_node)
                    if next_node.type == 'GROUP':
                        # Include nodes inside the group
                        if next_node.node_tree:
                            nodes_to_check.extend(next_node.node_tree.nodes)

        return False

    def set_custom_hdri_rotation(self, rotation_value):
        if isinstance(self.rotation_input.default_value, (mathutils.Euler, mathutils.Vector)):
            self.rotation_input.default_value = rotation_value
        elif isinstance(self.rotation_input.default_value, (float, int)):
            self.rotation_input.default_value = rotation_value
        else:
            self.report({'ERROR'}, "Unsupported rotation input type.")
            return
        # Force update of the scene to ensure the change takes effect
        bpy.context.view_layer.update()
        # Redraw the viewport
        self.safe_redraw()

    def set_mapping_node_rotation(self, rotation_value):
        if isinstance(self.mapping_node.inputs['Rotation'].default_value, mathutils.Vector):
            rotation = self.mapping_node.inputs['Rotation'].default_value.copy()
            if isinstance(rotation_value, mathutils.Vector):
                rotation[2] = rotation_value[2]
            else:
                rotation[2] = rotation_value  # Degrees
            self.mapping_node.inputs['Rotation'].default_value = rotation
        elif isinstance(self.mapping_node.inputs['Rotation'].default_value, mathutils.Euler):
            rotation = self.mapping_node.inputs['Rotation'].default_value.copy()
            if isinstance(rotation_value, mathutils.Euler):
                rotation.z = rotation_value.z
            else:
                rotation.z = rotation_value  # Degrees
            self.mapping_node.inputs['Rotation'].default_value = rotation
        elif isinstance(self.mapping_node.inputs['Rotation'].default_value, (float, int)):
            self.mapping_node.inputs['Rotation'].default_value = rotation_value  # Degrees
        else:
            self.report({'ERROR'}, "Unsupported rotation input type in mapping node.")
            return
        # Force update
        bpy.context.view_layer.update()
        self.safe_redraw()

    def get_builtin_hdr_rotation(self):
        # Get the built-in HDRI rotation in degrees
        rotation_rad = self.shading.studiolight_rotate_z
        rotation_deg = math.degrees(rotation_rad)
        return rotation_deg

    def set_builtin_hdr_rotation(self, rotation_value):
        # Set the built-in HDRI rotation
        rotation_rad = math.radians(rotation_value)
        # Wrap rotation to -pi to pi
        rotation_rad_wrapped = (rotation_rad + math.pi) % (2 * math.pi) - math.pi
        self.shading.studiolight_rotate_z = rotation_rad_wrapped
        self.safe_redraw()

    def calculate_new_direction(self, start_direction, rot_x, rot_z):
        # Rotate the start_direction vector around X and Z axes
        rot_x_mat = mathutils.Matrix.Rotation(rot_x, 4, 'X')
        rot_z_mat = mathutils.Matrix.Rotation(rot_z, 4, 'Z')

        # Apply rotations
        new_direction = rot_z_mat @ rot_x_mat @ start_direction

        # Normalize the vector
        if new_direction.length_squared != 0:
            new_direction.normalize()
        else:
            new_direction = mathutils.Vector((0.0, 0.0, -1.0))  # Default direction

        # Ensure the Z component is negative (light from above)
        if new_direction.z > 0:
            new_direction.z = -new_direction.z

        return new_direction

    def execute(self, context):
        self.cleanup()
        return {'FINISHED'}

    def cancel(self, context):
        # Restore original settings if necessary
        if self.mode == 'LIGHT_DIRECTION':
            self.scene.display.light_direction = self.start_light_direction
        elif self.mode == 'CUSTOM_HDRI_ROTATION':
            self.set_custom_hdri_rotation(self.start_rotation)
        elif self.mode == 'MAPPING_NODE_ROTATION':
            self.set_mapping_node_rotation(self.start_rotation)
        elif self.mode == 'HDRI_ROTATION':
            self.set_builtin_hdr_rotation(self.start_rotation)
        self.cleanup()
        return {'CANCELLED'}

def register():
    bpy.utils.register_class(HDRI_Rotator)

def unregister():
    bpy.utils.unregister_class(HDRI_Rotator)
