'''
HDRI Rotator Preferences
Author: 1P2D - Felix Pierron (me@1p2d.com)
Website: https://1p2d.com

This module defines the preferences for the HDRI Rotator addon.

License: GNU General Public License v3 (GPLv3)
'''

import bpy
import rna_keymap_ui
from . import bl_info

class HDRIRotatorPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__  # Use __package__ to refer to the addon module name

    sensitivity: bpy.props.FloatProperty(
        name="Sensitivity",
        description="Adjust the rotation sensitivity",
        default=0.8,
        min=0.1,
        max=1.0,
    )

    def draw(self, context):
        layout = self.layout

        # Draw sensitivity slider
        layout.prop(self, "sensitivity")

        # Draw Keymap Settings directly under Preferences
        self.draw_keymaps(context, layout)

    def draw_keymaps(self, context, layout):
        wm = context.window_manager
        kc = wm.keyconfigs.user

        # Keymap info specific to your addon
        keymaps_info = [
            {
                'km_name': '3D View',
                'space_type': 'VIEW_3D',
                'region_type': 'WINDOW',
                'operator_idname': 'view3d.hdri_rotator',
                'label': 'Keymap',
            },
        ]

        for km_info in keymaps_info:
            km = kc.keymaps.get(km_info['km_name'])
            if not km:
                continue

            for kmi in km.keymap_items:
                if kmi.idname == km_info['operator_idname']:
                    # Draw the keymap item without embedding it in an extra box
                    layout.label(text=km_info['label'])  # Label for the keymap
                    rna_keymap_ui.draw_kmi([], kc, km, kmi, layout, 0)
                    break

addon_keymaps = []

def register_keymap():
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        km = kc.keymaps.new(name='3D View', space_type='VIEW_3D')

        # Add the keymap for the operator
        kmi = km.keymap_items.new('view3d.hdri_rotator', 'RIGHTMOUSE', 'PRESS', alt=True, ctrl=True)
        kmi.active = True

        addon_keymaps.append((km, kmi))

def unregister_keymap():
    wm = bpy.context.window_manager
    kc = wm.keyconfigs.addon
    if kc:
        for km, kmi in addon_keymaps:
            km.keymap_items.remove(kmi)
        addon_keymaps.clear()

def register():
    bpy.utils.register_class(HDRIRotatorPreferences)
    register_keymap()

def unregister():
    unregister_keymap()
    bpy.utils.unregister_class(HDRIRotatorPreferences)
